//
// GCALDaemon is an OS-independent Java program that offers two-way
// synchronization between Google Calendar and various iCalalendar (RFC 2445)
// compatible calendar applications (Sunbird, Rainlendar, iCal, Lightning, etc).
//
// http://gcaldaemon.sourceforge.net
//
package org.gcaldaemon.core;

import java.util.Iterator;
import java.util.LinkedList;

import net.sf.gm.impl.GMClient;
import net.sf.gm.impl.GMLoginInfo;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Gmail connection pool.
 * 
 * Created: Jan 03, 2007 12:50:56 PM
 * 
 * @author Andras Berkes
 */
public final class GmailPool extends Thread {

	// --- CONSTANTS ---

	private static final long GMAIL_CONNECTION_TIMEOUT = (1000L * 60 * 5)
			- (1000L * 20);

	// --- LOGGER ---

	private static final Log log = LogFactory.getLog(GmailPool.class);

	// --- GMAIL CLIENT POOL ---

	private final LinkedList pool = new LinkedList();

	// --- CONSTRUCTOR ---

	public GmailPool(ThreadGroup mainGroup, Configurator configurator) {
		super(mainGroup, "Gmail pool");
		setPriority(Thread.NORM_PRIORITY - 2);
		start();
	}

	// --- BORROW OBJECT ---

	public final synchronized GmailEntry borrow(String username, String password)
			throws Exception {

		// Find entry
		long now = System.currentTimeMillis();
		Iterator entries = pool.iterator();
		GmailEntry entry;
		while (entries.hasNext()) {
			entry = (GmailEntry) entries.next();
			if (entry.username.equals(username)) {
				entries.remove();
				if (isTimeouted(now, entry)) {
					disconnect(entry);
				} else {
					log.debug("Gmail connection borrowed from the pool.");
					entry.lastUsage = now;
					return entry;
				}
			}
		}
		entry = new GmailEntry();
		entry.client = connect(username, password);
		entry.username = username;
		entry.lastUsage = now;
		return entry;
	}

	private static final GMClient connect(String username, String password)
			throws Exception {

		// Create Gmail connector
		log.debug("Connecting to Gmail...");
		GMLoginInfo login = new GMLoginInfo();
		login.setUsername(GCalUtilities.normalizeUsername(username));
		login.setPassword(password);
		String proxyHost = System.getProperty("http.proxyHost", "");
		String proxyPort = System.getProperty("http.proxyPort", "");
		if (proxyHost.length() != 0 && proxyPort.length() != 0) {
			login.setProxy(proxyHost);
			login.setProxyPort(Integer.parseInt(proxyPort));
			String proxyUsername = System.getProperty("http.proxyUserName", "");
			String proxyPassword = System.getProperty("http.proxyPassword", "");
			if (proxyUsername.length() != 0 && proxyPassword.length() != 0) {
				login.setProxyUserInfo(proxyUsername, proxyPassword);
			}
		}
		GMClient client = new GMClient(login);
		client.connect();
		if (client.isConnected()) {
			log.debug("Gmail connected successfully.");
		}
		return client;
	}

	// --- RECYCLE OBJECT ---

	public final synchronized void recycle(GmailEntry entry) {
		long now = System.currentTimeMillis();
		if (isTimeouted(now, entry)) {
			disconnect(entry);
			return;
		}
		if (entry.client != null) {
			try {
				if (entry.client.isConnected()) {
					log.debug("Gmail connection released.");
					pool.addFirst(entry);
				}
			} catch (Exception anyException) {
				log.debug(anyException);
			}
		}
	}

	// --- POOL CLEANUP ---

	public final void run() {
		for (;;) {
			try {

				// Check timeouts
				long now = System.currentTimeMillis();
				GmailEntry entry;
				synchronized (this) {
					if (!pool.isEmpty()) {
						Iterator entries = pool.iterator();
						while (entries.hasNext()) {
							entry = (GmailEntry) entries.next();
							if (isTimeouted(now, entry)) {
								entries.remove();
								disconnect(entry);
							}
						}
					}
				}

				// Wait
				sleep(10000);

			} catch (InterruptedException interrupt) {
				return;
			} catch (Exception poolException) {
				log.warn("Unexpected pooling error!", poolException);
				try {
					sleep(10000);
				} catch (Exception ignored) {
					return;
				}
			}
		}
	}

	private static final boolean isTimeouted(long now, GmailEntry entry) {
		if (entry == null) {
			return true;
		}
		return now - entry.lastUsage >= GMAIL_CONNECTION_TIMEOUT;
	}

	private static final void disconnect(GmailEntry entry) {
		if (entry != null && entry.client != null) {
			try {
				entry.client.disconnect();
			} catch (Exception ignored) {
			}
			log.debug("Gmail connection closed.");
		}
	}

	// --- STOP POOL ---

	public final void interrupt() {

		// Interrupt thread
		super.interrupt();

		// Disconnect all connections
		synchronized (this) {
			Iterator entries = pool.iterator();
			while (entries.hasNext()) {
				disconnect((GmailEntry) entries.next());
			}
			pool.clear();
		}
	}

}
